/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

/*-----------------------------------------------------------------------------*
 // include
 *-----------------------------------------------------------------------------*/
#include <float.h>
#include "harktool.h"

// extern
extern struct GLOBAL g;
extern bool hasParsedOpt;

void helpForCalcTfGeo(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;
    printf("\n");
    printf("\n");
    printf("Function:\n");
    printf("  calctfgeo\n");
    printf("\n");
    printf("Description:\n");
    printf("  Generate transfer function by geometric calculation\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("\n");
    exit(exitCode);
}

/*-----------------------------------------------------------------------------*
 * 幾何計算から伝達関数作成処理
 *-----------------------------------------------------------------------------*/
int harktoolcli_calctfgeo(int argc, char *argv[]) {
    char* tspList = NULL;
    char* micList = NULL;
    char* outputType = "LOC_SEP";
    enum TFType outputTFType = LOC_SEP;
    char* outputFile = NULL;
    int directLength = 32;
    int normalizeSrc = 0;
    int normalizeMic = 0;
    int normalizeFreq = 0;
    int resetMic = false;
    char* logLevel = "E";
    struct poptOption optionsTable[] =
            {
            /*{ "longname", "shortname",
             *  argInfo,
             *  *arg, retval,
             *  description, argment description} */
            { "tsp-list", '\0',
              POPT_ARG_STRING,
              &tspList, 'i',
              "TSP list file  (ex. tsp.xml)", NULL },
            { "mic-list", '\0',
              POPT_ARG_STRING,
              &micList, 'm',
              "Microphone list file  (ex. microphone.xml)", NULL },
            { "output-type", '\0',
              POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
              &outputType, 't',
              "Transfer function type  LOC=Localization,SEP=Separation,LOC_SEP=LOC&SEP",
              "LOC|SEP|LOC_SEP" },
            { "output-file", '\0',
              POPT_ARG_STRING,
              &outputFile, 'o',
              "Output transfer function zip file path  (ex. /path/to/transfer.zip)", NULL },
            { "direct-length", '\0',
              POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
              &directLength, 'd',
              "Direct sound length", NULL },
            { "normalize-src", '\0',
              POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
              &normalizeSrc, 'S',
              "Normalize by src  0=disable,1=enable", "0|1" },
            { "normalize-mic", '\0',
              POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
              &normalizeMic, 'M',
              "Normalize by mic  0=disable,1=enable", "0|1" },
            { "normalize-freq", '\0',
              POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
              &normalizeFreq, 'F',
              "Normalize by frequency  0=disable,1=enable", "0|1" },
            { "reset-mic", '\0',
              POPT_ARG_NONE,
              &resetMic, 'R',
              "Remove unused channels from microphones.xml and reset their ids. (for HARKTOOL4 compatible)", NULL },
            { "logLevel", '\0',
              POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
              &logLevel, '-',
              "Log information level", "E|W|I|D" },
            { "help", '\0',
              0,
              NULL, '?',
              "Show this help message", NULL },
            POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**) argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "calctfgeo [OPTIONS]*");

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case 't':
            if (strcmp(outputType, "LOC") == 0 || strcmp(outputType, "loc") == 0) {
                outputTFType = LOC;
            } else if (strcmp(outputType, "SEP") == 0 || strcmp(outputType, "sep") == 0) {
                outputTFType = SEP;
            } else if (strcmp(outputType, "LOC_SEP") == 0 || strcmp(outputType, "loc_sep") == 0) {
                outputTFType = LOC_SEP;
            } else {
                harkio_Log_printf(harkio_Log_Error, "%s: %s", poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
                        "invalid string value");
                exit(EXIT_FAILURE);
            }
            hasParsedOpt = true;
            break;
        case 'd':
            if (directLength < 0) {
                c = POPT_ERROR_BADNUMBER;
                goto OPTIONERR;
            }
            hasParsedOpt = true;
            break;
        case 'S':
            if (normalizeSrc != 0 && normalizeSrc != 1) {
                c = POPT_ERROR_BADNUMBER;
                goto OPTIONERR;
            }
            hasParsedOpt = true;
            break;
        case 'M':
            if (normalizeMic != 0 && normalizeMic != 1) {
                c = POPT_ERROR_BADNUMBER;
                goto OPTIONERR;
            }
            hasParsedOpt = true;
            break;
        case 'F':
            if (normalizeFreq != 0 && normalizeFreq != 1) {
                c = POPT_ERROR_BADNUMBER;
                goto OPTIONERR;
            }
            hasParsedOpt = true;
            break;
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            helpForCalcTfGeo(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        OPTIONERR:
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s", poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    // check essential option
    bool isOptOK = true;
    if (!tspList) {
        harkio_Log_printf(harkio_Log_Error, "missing --tsp-list");
        isOptOK = false;
    }
    if (!micList) {
        harkio_Log_printf(harkio_Log_Error, "missing --mic-list");
        isOptOK = false;
    }
    if (!outputFile) {
        harkio_Log_printf(harkio_Log_Error, "missing --output-file");
        isOptOK = false;
    }
    if (!isOptOK)
        return EXIT_FAILURE;

    /*--------------------------------------------------*/
    //伝達関数作成
    /*--------------------------------------------------*/
    harkio_Log_printf(harkio_Log_Info, "@ harktoolcli_calctfgeoExec start");
    int ret = harktoolcli_calctfgeoExec(tspList, micList, outputTFType, outputFile, directLength, normalizeSrc,
            normalizeMic, normalizeFreq, resetMic);

    /*--------------------------------------------------*/
    //* end proc
    /*--------------------------------------------------*/
    harkio_Log_printf(harkio_Log_Info, "@ harktoolcli_calctfgeoExec %s end", (ret == EXIT_SUCCESS)?"normal":"error");

    // free poptContext
    poptFreeContext(optCon);

    harktool_Exit(ret);

    return ret;
} //end

void writeNormalizeInfo(int normSrc, int normMic, int normFreq) {
    // write normalization = 1 to README
    cat_asprintf(&g.readmeMsg, "%s \n%s[%d]\n%s[%d]\n%s[%d]\n", "Normalization Type", "- Source     ", 1,
            "- Microphone ", 1, "- Frequency  ", 1);
    if (normSrc == 0 || normMic == 0 || normFreq == 0) {
        harkio_Log_printf(harkio_Log_Warning,
                "ignored --normalize*=0 option. calctfgeo always generates normalized value.");
    }
}

/*-----------------------------------------------------------------------------*
 * 幾何計算から伝達関数作成処理
 *-----------------------------------------------------------------------------*/
int harktoolcli_calctfgeoExec(const char* tspFile, const char* micFile, enum TFType outputType, const char* outputFile,
        int directLength, int normSrc, int normMic, int normFreq, bool resetMic) {
    g.readmeMsg = NULL;
    /*--------------------------------------------------*/
    // tsp-listファイル(source.xml)->XML構造体へ
    /*--------------------------------------------------*/
    g.src_xml = harkio_XML_newFromXML(tspFile, "");
    if (g.src_xml == NULL) {
        harkio_Log_FileOpenFailed(tspFile);
        return EXIT_FAILURE;
    }

    /*--------------------------------------------------*/
    // micファイル(microphones.xml)->XML構造体へ
    /*--------------------------------------------------*/
    g.mic_xml = harkio_XML_newFromXML(micFile, "");
    if (g.mic_xml == NULL) {
        harkio_Log_FileOpenFailed(micFile);
        return EXIT_FAILURE;
    }

    /*--------------------------------------------------*/
    // position type
    /*--------------------------------------------------*/
    g.PosType = TSP;

    /*--------------------------------------------------*/
    //mic数
    /*--------------------------------------------------*/
    g.Nmic = harkio_Positions_getSize(g.mic_xml->poses);

    /*--------------------------------------------------*/
    //音源数
    /*--------------------------------------------------*/
    g.Nsrc = harkio_Positions_getSize(g.src_xml->poses);

    // allocate and init output channels
    bool outputChannels[MAX_CH] = { }; // init false
    if (g.Nmic > 0 && g.Nsrc > 0 && g.src_xml->poses->pos[0]->channelsUse == 1) {
        if (resetMic) {
            // write title
            cat_asprintf(&g.readmeMsg, "Original Microphone Channels ");
        }
        for (int i = 0; i < g.src_xml->poses->pos[0]->numChannels; i++) {
            if (g.src_xml->poses->pos[0]->channels[i] < MAX_CH) {
                harkio_Log_printf(harkio_Log_Debug, "use channel %d", g.src_xml->poses->pos[0]->channels[i]);
                outputChannels[g.src_xml->poses->pos[0]->channels[i]] = true;
                if (resetMic) {
                    // write channels as CSV
                    cat_asprintf(&g.readmeMsg, "%d,", g.src_xml->poses->pos[0]->channels[i]);
                }
            }
        }
        if (resetMic) {
            // last comma replace \n
            g.readmeMsg[strlen(g.readmeMsg) - 1] = '\n';
        }
        g.Nmic = g.src_xml->poses->pos[0]->numChannels;
    } else {
        harkio_Log_printf(harkio_Log_Debug, "use all channels. NumCh=%d", g.Nmic);
        memset(outputChannels, true, sizeof(bool) * g.Nmic);
    }

    /*--------------------------------------------------*/
    // Config取得
    /*--------------------------------------------------*/
    if (g.src_xml->cfg == NULL) {
        harkio_Log_printf(harkio_Log_Error, "Config not found in source.xml");
        return EXIT_FAILURE;
    }
    g.Nfft = harkio_Config_getNfft(g.src_xml->cfg);
    g.TspLength = harkio_Config_getTSPLength(g.src_xml->cfg);
    g.SamplingRate = harkio_Config_getSamplingRate(g.src_xml->cfg);
    g.SignalMax = harkio_Config_getSignalMax(g.src_xml->cfg);
    g.SyncAverage = harkio_Config_getSynchronousAverage(g.src_xml->cfg);
    g.OrgTspPath = harkio_Config_getTSPpath(g.src_xml->cfg);
    g.TspOffset = harkio_Config_getTSPoffset(g.src_xml->cfg);
    g.PeakSearchFrom = harkio_Config_getPeakSearchFrom(g.src_xml->cfg);
    g.PeakSearchTo = harkio_Config_getPeakSearchTo(g.src_xml->cfg);

    g.IRsize = g.Nsrc * g.Nmic * g.Nfft;
    g.TFsize = g.Nsrc * g.Nmic * (g.Nfft / 2 + 1);

    /*--------------------------------------------------*/
    /* 領域確保                                         */
    /*--------------------------------------------------*/
    // 伝達関数用領域
    // npos(72) x Nmic(8) x Nfft(512)/2 + 1 x sizeof(complex) 分の領域確保
    harktool_new(TF_VECT);

    // 伝達関数 作業用領域
    // npos(72) x Nmic(8) x Nfft(512)/2 + 1 x sizeof(complex) 分の作業領域確保
    harktool_new(TMP_TF_VECT);

    // 伝達関数作成用TEMP領域
    harktool_new(BUFFER);

    // MIC position
    harktool_new(MIC_POS_VECT);

    // SRC position
    harktool_new(SRC_POS_VECT);

    // SORT
    harktool_new(SORT_POS_ID);

    // インパルス応答ファイルのPATH格納領域
    harktool_new(FLT_PATH);

    // TSP wavファイル名格納領域
    harktool_new(TSP_WAV_NAME);

    /*--------------------------------------------------*/
    /*  Create Directory                                */
    /*--------------------------------------------------*/
    harktool_createTmpPath();

    /*--------------------------------------------------*/
    // id を昇順に並べ替えてpositionをセット
    /*--------------------------------------------------*/
    // MIC ID 設定
    harktool_setMicPosId(outputChannels);
    // mic positionセット + 極座標->直交座標変換処理
    harktool_setMicPos();
    // SRC ID 設定
    harktool_setSrcPosId();
    // tsp positionセット + 極座標->直交座標変換処理
    harktool_setSrcPos();

    // allocate position info
    SRC_POS_INFO posInfoInit = {{0.0, 0, 0},{0.0, 0, 0}}, srcPosInfo[g.Nsrc];
    for (int i = 0; i < g.Nsrc; i++) {
        srcPosInfo[i] = posInfoInit;
    }

    /*--------------------------------------------------*/
    // 計算処理
    /*--------------------------------------------------*/

//    //mic positionの平均
//    harktool_calcMicPosMean();
//
//    for (int i = 0; i < 3; i++) {
//        harkio_Log_printf(harkio_Log_Debug, "PosMean[%d]=%f", i, g.PosMean[i]);
//        for (int mic = 0; mic < g.Nmic; mic++) {
//            g.MicPosVect[mic][i] -= g.PosMean[i];
//        }
//    }
    const float sound_speed = SOUND_SPEED;

    // norm_2 v = sqrt (sum (v [i] * v [i]))

    for (int src = 0; src < g.Nsrc; src++) {
        printf(" progress:%d / %d\n", src+1, g.Nsrc);
        fflush(stdout);

        float maxDelta = 0;
        //int maxDeltaCh = 0;
        float srcDelta[g.Nmic];
        for (int mic = 0; mic < g.Nmic; mic++) {
            float distance = 0.;
//          square = harktool_pow(g.SrcPosVect[src][0] - g.MicPosVect[mic][0], 2)
//                    + harktool_pow(g.SrcPosVect[src][1] - g.MicPosVect[mic][1], 2)
//                    + harktool_pow(g.SrcPosVect[src][2] - g.MicPosVect[mic][2], 2);
            for (int i = 0; i < 3; i++) {
                distance += powf(g.SrcPosVect[src][i] - g.MicPosVect[mic][i], 2);
            }
            srcDelta[mic] = g.SamplingRate / sound_speed * sqrtf(distance);

            harkio_Log_printf(harkio_Log_Debug, "mic [%d] : norm_2=%f : srcDelta=%f", g.MicPosId[mic],
                    sqrtf(distance), srcDelta[mic]);
            if (maxDelta < srcDelta[mic]) {
                maxDelta = srcDelta[mic];
                //maxDeltaCh = mic;
            }
        }

        float offset = maxDelta + directLength - g.Nfft;
        if (offset < 0)
            offset = 0;
        srcPosInfo[src].loc.tfOffset = srcPosInfo[src].sep.tfOffset = offset;
        harkio_Log_printf(harkio_Log_Debug, "GEO IR[%d] offset=%f(%fsec)", g.SrcPosId[src], offset, offset / 16000.0);

        for (int mic = 0; mic < g.Nmic; mic++) {
//            if (srcDelta[mic] >= offset) {
                // mic within range
                for (int fft = 0; fft < g.Nfft / 2 + 1; fft++) {
                    float theta = 2.0 * pi * (srcDelta[mic] - offset) * fft / g.Nfft;
                    g.TFvect[src][mic][fft] = cexpf(-theta * _Complex_I);//cos(theta) + sin(theta) * _Complex_I;
                }
//            } else {
//                // mic out of range
//                harkio_Log_printf(harkio_Log_Warning,
//                        "Cannot generate TF. because distance too far between mic[%d] and mic[%d] on src[%d] or direct-length(%d) is too large.",
//                        g.MicPosId[maxDeltaCh], g.MicPosId[mic], g.SrcPosId[src], directLength);
//                cat_asprintf(&g.readmeMsg,
//                        "Cannot generate TF. because distance too far between mic[%d] and mic[%d] on src[%d] or direct-length(%d) is too large.\n",
//                        g.MicPosId[maxDeltaCh], g.MicPosId[mic], g.SrcPosId[src], directLength);
//                for (int fft = 0; fft < g.Nfft / 2 + 1; fft++) {
//                    g.TFvect[src][mic][fft] = 0.0;
//                }
//            }
        }
    }

    switch (outputType) { //outputType:SEP,LOC,LOC_SEP
    case LOC:
        //定位伝達関数の配列からharkio_Matrix型に変換する
        harktool_setLoctfFromArray();
        break;
    case SEP:
        writeNormalizeInfo(normSrc, normMic, normFreq);
        harktool_conjTF();
        //分離伝達関数の配列からharkio_Matrix型に変換する
        harktool_setSeptfFromArray();
        break;
    case LOC_SEP:
        harktool_setLoctfFromArray();
        writeNormalizeInfo(normSrc, normMic, normFreq);
        harktool_conjTF(); // NOTE: TFvect replaced
        harktool_setSeptfFromArray();
        break;
    }

    // write sound info to README
    cat_asprintf(&g.readmeMsg, "Direct sound Length  %d\n", directLength);

    /*--------------------------------------------------*/
    // 伝達関数をZipファイルに出力
    /*--------------------------------------------------*/
    writeTF(outputType, outputFile, resetMic);

    /*--------------------------------------------------*/
    // READMEに設定情報を書き込む
    /*--------------------------------------------------*/
    appendInfoToZipFile(outputFile, srcPosInfo);

    return EXIT_SUCCESS;
} //main end

