/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "harktool.h"

#ifdef WIN32
#define NULL_DEVICE "NUL"
#else
#define NULL_DEVICE "/dev/null"
#endif

#ifdef WIN32
int vscprintf(const char *format, va_list ap) {
    int iResult = -1;
    FILE *fp = fopen(NULL_DEVICE, "w");
    if (fp) {
        iResult = vfprintf(fp, format, ap);
        fclose(fp);
    }

    return iResult;
}

int vasprintf(char **strp, const char *format, va_list ap) {
    int iResult = -1;
    int iLength = vscprintf(format, ap);
    if (iLength >= 0) {
        char *str = malloc(sizeof(char) * (iLength + 1));
        if (str) {
            iResult = vsprintf(str, format, ap);
            if (iResult >= 0) {
                *strp = str;
                return iResult;
            }

            free(str);
        }
    }

    return iResult;
}

int asprintf(char **strp, const char *format, ...) {
    int iResult;
    va_list ap;

    va_start(ap, format);
    iResult = vasprintf(strp, format, ap);
    va_end(ap);

    return iResult;
}
#endif

// *strp += asprintf()
int cat_asprintf(char** strp, const char* format, ...) {
    if (strp == NULL)
        return 0;
    char* addStr;
    va_list ap;
    va_start(ap, format);
    int addLen = vasprintf(&addStr, format, ap);
    va_end(ap);
    if (addLen < 0)
        return addLen;

    if (*strp == NULL) {
        // return addStr only
        *strp = addStr;
        return addLen;
    }

    // cat str + addStr
    char* newStr = realloc(*strp, strlen(*strp) + addLen + 1);
    if (newStr == NULL) {
        harkio_Log_printf(harkio_Log_Error, "realloc error");
        free(addStr);
        return -1;
    }
    *strp = newStr;

    strcat(*strp, addStr);
    free(addStr);
    return strlen(*strp);
}

int as_zip_fread(struct zip* handle, const char* srcPath, char** pStrBuf) {
    struct zip_stat zs;
    zip_stat_init(&zs);
    if (zip_stat(handle, srcPath, ZIP_STAT_SIZE, &zs) != 0) {
        harkio_Log_printf(harkio_Log_Error, "zip_stat error (%s)",
                zip_strerror(handle));
        return EXIT_FAILURE;
    }
    *pStrBuf = (char*) calloc(zs.size + 1, sizeof(char));
    if (*pStrBuf == NULL) {
        harkio_Log_printf(harkio_Log_Error, "memory allocate error");
        return EXIT_FAILURE;
    }
    struct zip_file* zf = zip_fopen(handle, srcPath, 0);
    if (zf == NULL) {
        harkio_Log_printf(harkio_Log_Error, "zip_fopen error (%s)",
                zip_strerror(handle));
        free(*pStrBuf);
        return EXIT_FAILURE;
    }
    if (zip_fread(zf, *pStrBuf, zs.size) == -1) {
        harkio_Log_printf(harkio_Log_Error, "zip_fread error (%s)",
                zip_file_strerror(zf));
        zip_fclose(zf);
        return EXIT_FAILURE;
    }
    zip_fclose(zf);
    return EXIT_SUCCESS;
}
