/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <math.h>
#include <fftw.h>
#include <rfftw.h>
#include <string.h>

//#define DEBUG_CONV
//#define DEBUG_CONV1
//#define DEBUG_CONV2

void setxi( double *x, int nx, int m, double *xi );
void sethi( double *h, int n2, int m, int nh, int index, double *hi );
int as_fft( double *in, int nfft, rfftwnd_plan *p, fftw_complex *out );
int as_ifft( fftw_complex *in, int nfft, rfftwnd_plan *p, double *out );

int addwave( double *hi, int n1, int n2, double *wave, int index, int maxi );
void multiply( int n, fftw_complex *in1, fftw_complex *in2, fftw_complex *out );
void divide( int n, fftw_complex *in1, fftw_complex *in2, fftw_complex *out );
int cutwave( double *hi, int n1, int n2, double *wave, int index, int maxi );
fftw_complex *mallocSpec( int length );


void printdoubles(int m, double *wave){
  int j;

  for( j = 0; j < m; j ++ ){
    printf("%f ", wave[j]);
  }
  printf("\n");
}

void printcomplexes( int m, fftw_complex *wave ){
  int j;

  for( j = 0; j < m; j ++ ){
    printf("%f+%fi ", c_re(wave[j]), c_im(wave[j]) );
  }
  printf("\n");
}

fftw_complex *mallocSpec( int length ){
  fftw_complex *data;

  if( ( data = (fftw_complex *)malloc( sizeof(fftw_complex) * length ) ) == NULL ){
    perror("malloc");
    return NULL;
  }
  return data;
}

/* y(t) = h(t) * x(t) */
int convolusion(double *h, int nh, double *x, int nx, double **wave){

  double *hi;
  double *xi;
  fftw_complex *hisp, *xsp;
  rfftwnd_plan fftp, ifftp;
  int i, j;
  int len;
  
  int n1, n2;
  int m;
  int mm;

  n1 = nx; 
  n2 = nx * 2;  /* This is not optimal length... ad hoc... */
  mm = (int)ceil(log( n1 + n2 - 1 )/log(2));
  m = (int)pow(2, mm); /* for fft */

#ifdef DEBUG_CONV1
  printf( "n1=%d n2=%d m=%d\n", n1, n2, m );
#endif

  if( ( hi = (double *)malloc( sizeof(double) * m ) ) == NULL ){
    perror("malloc");
    exit(1);
  }

  if( ( *wave = (double *)malloc( sizeof(double) * (nx + nh - 1) ) ) == NULL ){
    perror("malloc");
    exit(1);
  }

  if( ( xi = (double *)malloc( sizeof(double) * m ) ) == NULL ){
    perror("malloc");
    exit(1);
  }

  if( (hisp = mallocSpec( m )) == NULL ){
    perror("malloc");
    exit(1);
  }
    
  if( (xsp = mallocSpec( m )) == NULL ){
    perror("malloc");
    exit(1);
  }

  fftp = rfftwnd_create_plan( 1, &m, FFTW_FORWARD, FFTW_ESTIMATE );
  ifftp = rfftwnd_create_plan( 1, &m, FFTW_BACKWARD, FFTW_ESTIMATE );

  setxi( x, nx, m, xi );

#ifdef DEBUG_CONV1
  printf("xi=");
  printdoubles( m, xi );
#endif

  as_fft( xi, m, &fftp, xsp );

#ifdef DEBUG_CONV1
  printf("xsp=");
  printcomplexes( m, xsp );
#endif

  for( i = -n1+1, j = 0; i < nh; i += (n2-n1+1), j++ ){
    sethi( h, n2, m, nh, i, hi );

#ifdef DEBUG_CONV1
    printf("hi=");
    printdoubles( m, hi );
#endif

    as_fft( hi, m, &fftp, hisp );

#ifdef DEBUG_CONV1
    printf("hisp=");
    printcomplexes( m, hisp );
#endif

    multiply( m, hisp, xsp, hisp );

#ifdef DEBUG_CONV1
    printf("res=");
    printcomplexes( m, hisp );
#endif

    as_ifft(hisp, m, &ifftp, hi );

#ifdef DEBUG_CONV1
    printf("ifft=");
    printdoubles( m, hi );
#endif

    len = cutwave( hi, n1, n2, *wave, (n2-n1+1)*j, nh+nx-1 );

#ifdef DEBUG_CONV1
    printf( "%d cut\n", len );
    printf("wave=");
    printdoubles( nx+nh-1, *wave );
#endif

  }

  free(hi);
  free(xi);
  free(hisp);
  free(xsp);
  rfftwnd_destroy_plan(fftp);  
  rfftwnd_destroy_plan(ifftp);  

  return (nx + nh - 1);
}

/* y(t) = h(t) * x(t) */
int deconvolusion(double *h, int nh, double *x, int nx, double **wave){

  double *hi;
  double *xi;
  fftw_complex *hisp, *xsp;
  rfftwnd_plan fftp, ifftp;
  int i, j;
  int len;
  
  int n1, n2;
  int m;
  int mm;

  n1 = nx; 
  n2 = nx * 2;  /* This is not optimal length... ad hoc... */
  mm = (int)ceil(log( n1 + n2 - 1 )/log(2));
  m = (int)pow(2, mm); /* for fft */

#ifdef DEBUG_CONV2
  printf( "n1=%d n2=%d m=%d\n", n1, n2, m );
#endif

  if( ( hi = (double *)malloc( sizeof(double) * m ) ) == NULL ){
    perror("malloc");
    exit(1);
  }

  if( ( *wave = (double *)malloc( sizeof(double) * (nh - nx + 1) ) ) == NULL ){
    perror("malloc");
    exit(1);
  }
  bzero( *wave, sizeof(double) * (nh - nx + 1) );

  if( ( xi = (double *)malloc( sizeof(double) * m ) ) == NULL ){
    perror("malloc");
    exit(1);
  }

  if( (hisp = mallocSpec( m )) == NULL ){
    perror("malloc");
    exit(1);
  }
    
  if( (xsp = mallocSpec( m )) == NULL ){
    perror("malloc");
    exit(1);
  }

  fftp = rfftwnd_create_plan( 1, &m, FFTW_FORWARD, FFTW_ESTIMATE );
  ifftp = rfftwnd_create_plan( 1, &m, FFTW_BACKWARD, FFTW_ESTIMATE );

  setxi( x, nx, m, xi );

#ifdef DEBUG_CONV2
  printf("xi=");
  printdoubles( m, xi );
#endif

  as_fft( xi, m, &fftp, xsp );

#ifdef DEBUG_CONV2
  printf("xsp=");
  printcomplexes( m, xsp );
#endif

  for( i = 0, j = 0; i < nh; i += (n2-n1+1), j++ ){
    setxi( &h[i], n2, m, hi );

#ifdef DEBUG_CONV2
    printf("hi=");
    printdoubles( m, hi );
#endif

    as_fft( hi, m, &fftp, hisp );

#ifdef DEBUG_CONV2
    printf("hisp=");
    printcomplexes( m, hisp );
#endif

    divide( m, hisp, xsp, hisp );

#ifdef DEBUG_CONV2
    printf("res=");
    printcomplexes( m, hisp );
#endif

    as_ifft(hisp, m, &ifftp, hi );

#ifdef DEBUG_CONV2
    printf("ifft=");
    printdoubles( m, hi );
#endif

    len = addwave( hi, n1, n2, *wave, (n2-n1+1)*j, nh-nx+1 );

#ifdef DEBUG_CONV2
    printf( "%d add\n", len );
    printf("wave=");
    printdoubles( nh-nx+1, *wave );
#endif

  }

  free(hi);
  free(xi);
  free(hisp);
  free(xsp);
  rfftwnd_destroy_plan(fftp);  
  rfftwnd_destroy_plan(ifftp);  

  return (nh - nx + 1);
}

int cutwave( double *hi, int n1, int n2, double *wave, int index, int maxi ){
  int i, j;

  for( i = n1-1, j = 0; i < n2; i++, j++ ){
    if(index +j < maxi ){
      wave[index + j] = hi[i];
    } else break;
  }

  return j;
}

int addwave( double *hi, int n1, int n2, double *wave, int index, int maxi ){
  int j;

  for( j = 0; j < n2; j++ ){
    if(index +j < maxi ){
      wave[index + j] += hi[j];
    } else break;
  }

  return j;
}

void multiply( int n, fftw_complex *in1, fftw_complex *in2, fftw_complex *out ){
  fftw_complex tmp;
  int i;
  
  for( i = 0; i < n; i++ ){
    c_re(tmp) =  c_re(in1[i]) * c_re(in2[i]) - c_im(in1[i]) * c_im(in2[i]);
    c_im(tmp) =  c_re(in1[i]) * c_im(in2[i]) + c_im(in1[i]) * c_re(in2[i]);
    c_re(out[i]) = c_re(tmp);
    c_im(out[i]) = c_im(tmp);
  }
}

void divide( int n, fftw_complex *in1, fftw_complex *in2, fftw_complex *out ){
  fftw_complex tmp;
  double d;
  int i;
  
  for( i = 0; i < n; i++ ){
    d = c_re(in2[i]) * c_re(in2[i]) + c_im(in2[i]) * c_im(in2[i]);
    if( d == 0 ){
      perror("divide");
      c_re(out[i]) = c_im(out[i]) = 0;
    } else {
      c_re(tmp) =  c_re(in1[i]) * c_re(in2[i]) + c_im(in1[i]) * c_im(in2[i]);
      c_im(tmp) =  - c_re(in1[i]) * c_im(in2[i]) + c_im(in1[i]) * c_re(in2[i]);
      c_re(out[i]) = c_re(tmp)/d;
      c_im(out[i]) = c_im(tmp)/d;
    }
  }
}

void setxi( double *x, int nx, int m, double *xi ){
  int i;

  for( i = 0; i < m; i++ ){
    if( i < nx )
      xi[i] = x[i];
    else
      xi[i] = 0;
  }
}

void sethi( double *h, int n2, int m, int nh, int index, double *hi ){
  int i;

  for( i = 0; i < m; i++ ){
    if( i < n2 && index + i < nh && index + i >= 0){
      hi[i] = h[index + i];
    } else {
      hi[i] = 0.0;
    }
  }
}

int as_fft(double *in, int nfft, rfftwnd_plan *p, fftw_complex *out ){

  int i;
  fftw_real *in2;

  if( (in2 = (fftw_real *)malloc(sizeof(fftw_real)* nfft )) == NULL ){
    perror("malloc");
    return -1;
  }

  for( i = 0; i < nfft; i++ ){
    //in2[i] = (fftw_real)in[i] * (0.5 - 0.5 * cos( 2*M_PI / nfft * i ));
    in2[i] = (fftw_real)in[i];
  }

  rfftwnd_one_real_to_complex(*p, in2, out);

  free( in2 );

  return 1;
}

int as_ifft( fftw_complex *in, int nfft, rfftwnd_plan *p, double *out ){

  fftw_real *out2;
  int i;

  if( (out2 = (fftw_real *)malloc(sizeof(fftw_real)* nfft )) == NULL ){
    perror("malloc");
    return -1;
  }

  rfftwnd_one_complex_to_real(*p, in, out2);

  for( i = 0; i < nfft; i++){
    //    out[i] = (double)(out2[i]/(0.5 - 0.5 * cos( 2*M_PI / nfft * i )) / nfft ); 
    out[i] = (double)(out2[i]/nfft); 
  }

  free(out2);
  return 1;
}

#ifdef DEBUG_CONV
int main(){
  double h[] = { 1,2,3,4,5,6,7,8 };
  double x[] = { 1,1 };
  double *res;
  double *res2;
  int len, len2;

  
  printf("[");
  printdoubles( 8, h );
  printf("] * [");
  printdoubles( 2, x );
  printf("] =\n");

  len = convolusion(h, 8, x, 2, &res);
  printf("wave=");
  printdoubles( len, res );

  len2 = deconvolusion(res, len, x, 2, &res2);
  printf("wave=");
  printdoubles( len2, res2 );

  free(res);
  free(res2);
}
#endif
