/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <popt.h>
#include <math.h>
#include <libharkio3.h>

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-conv-sm\n");
    printf("\n");
    printf("Description:\n");
    printf("  Convert the old file separation matrix\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("Arguments:\n");
    printf("  sepmatfile  Separation matrix file               (e.g., mat.tff)\n");
    printf("  septffile   Corresponding transfer function file (e.g., sep.zip)\n");
    printf("  outputfile  Output separation matrix file        (e.g., out.zip)\n");
    printf("\n");
    printf("\n");
    exit(exitCode);
}

int main(int argc, char **argv)
{
    char *whatisthis= "separation matrix";
    char* logLevel = "E";

    harkio_TransferFunction *tf = NULL;

    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* sepmatfile(binary) septffile(zip) outputfile(zip)");

    //argument check
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error , "argument error!!");
        help(optCon, EXIT_FAILURE);
    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    const char* sepmatpath = poptGetArg(optCon);
    if (!sepmatpath) {
        harkio_Log_printf(harkio_Log_Error , "missing sepmatfile(binary)");
    } else {
        printf("sepmatpath=%s\n",sepmatpath);
    }

    const char* septfpath = poptGetArg(optCon);
    if (!septfpath) {
        harkio_Log_printf(harkio_Log_Error , "missing septffile(zip)");
    } else {
        printf("septfpath=%s\n",septfpath);
    }

    if (!sepmatpath || !septfpath) {
        exit(EXIT_FAILURE);
    }

    // result = libharkio3_SeparationMatrix_readlibhakio2( &tf , sepmatpath, septfpath );
    tf = harkio_Compat_SeparationMatrix_readlibharkio2(sepmatpath, septfpath);

    if (tf == NULL) {
        harkio_TransferFunction_delete(&tf);
        harkio_Log_printf(harkio_Log_Error , "harkio_Compat_SeparationMatrix_readlibharkio2 error");
        exit(EXIT_FAILURE);
    }

    const char* outfile = poptGetArg(optCon);
    if (!outfile) {
        harkio_Log_printf(harkio_Log_Error , "missing outputfile(zip)");
        exit(EXIT_FAILURE);
    }

    //write zip file
    harkio_TransferFunction_write(tf , outfile , whatisthis);

    fprintf(stderr, "SUCCESS\n");
    fprintf(stderr, " output file : %s\n",outfile);

    harkio_TransferFunction_delete(&tf);

    // free poptContext
    poptFreeContext(optCon);

    exit(EXIT_SUCCESS);
}//func end


