/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdbool.h>
#include <popt.h>
#include <libharkio3.h>

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-genneighbors\n");
    printf("\n");
    printf("Description:\n");
    printf("  The command creates an xml file that contains\n");
    printf("  neighbors tag of xml file.\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("   -n and -t are options for -a 1\n");
    printf("\n");
    printf("\n");
    exit(exitCode);
}

int main(int argc, char **argv)
{
    harkio_XML *xml;
    int  ret;
    FILE *fp;
    harkio_Neighbors *nbrs;

    int   alg = 1;
    int   nearest = 3;
    float threshold = 0.5;

    char* logLevel = "E";
    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "algorithm", 'a',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &alg, 'a',
      "Algorithms (1:NearestNeighbor)", NULL },
    { "num-max", 'n',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &nearest, 'n',
      "Maximum number of neighbors", NULL },
    { "threshold", 't',
      POPT_ARG_FLOAT | POPT_ARGFLAG_SHOW_DEFAULT,
      &threshold, 't',
      "Threshold for neighbor decision", NULL },
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* input_filename [output_filename]");

    //argument check
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error , "argument error!!");
        help(optCon, EXIT_FAILURE);
    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    // get input filename
    const char* infile = poptGetArg(optCon);

    if (!infile) {
        harkio_Log_printf(harkio_Log_Error , "missing input_filename");
        exit(EXIT_FAILURE);
    }

    xml =  harkio_XML_newFromXML(infile , "");
    nbrs = harkio_Neighbors_new(xml->poses);

    switch (alg) {
    case NearestNeighbor:
        ret = harkio_Neighbors_calcNeighbors(nbrs, NearestNeighbor,  nearest, threshold);
        if (ret == EXIT_FAILURE) {
            harkio_Log_printf(harkio_Log_Error, "harkio_Neighbors_calcNeighbors error");
            exit(EXIT_FAILURE);
        }
        break;
    default:
        harkio_Log_printf(harkio_Log_Error, "unknown algorithm [%d]", alg);
        exit(EXIT_FAILURE);
    }

    //set neighbors
    xml->nbrs = nbrs;

    // get output filename
    const char* outfile = poptGetArg(optCon);

    if (outfile == NULL) {
        fp = stdout;
    } else if ((fp = fopen(outfile , "w")) == NULL) {
        harkio_Log_FileOpenFailed(outfile);
        exit(EXIT_FAILURE);
    }

    // make new XML object and write
    harkio_XML_write(xml, fp);
    harkio_XML_delete(&xml);

    if (outfile != NULL)
        fclose(fp);

    harkio_Log_printf(harkio_Log_Debug, "XML file created");

    fprintf(stderr, "SUCCESS\n");
    if (outfile != NULL)
        fprintf(stderr, " output file : %s\n", outfile);

    // free poptContext
    poptFreeContext(optCon);

    exit(EXIT_SUCCESS);
}
