/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <popt.h>
#include <math.h>
#include <sndfile.h>
#include <libharkio3.h>

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-gentsp\n");
    printf("\n");
    printf("Description:\n");
    printf("  TSP signal generating tool. The file is saved as 32bit float wave.\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("\n");
    exit(exitCode);
}

/*------------------------------------------------------*/
// read 16384.little_endian.tsp
/*------------------------------------------------------*/
float* readRawTSP(char *path, long *tsplen)
{
    FILE* fp = fopen(path, "rb");
    long size, retsize;
    float *tsp;

    if (fp == NULL) {
        harkio_Log_FileOpenFailed(path);
        return NULL;
    }

    // check file size
    fseek(fp, 0, SEEK_END);
    size = ftell(fp) / sizeof(float);
    rewind(fp);

    harkio_Log_printf(harkio_Log_Debug, "TSP file includes %ld samples.", size);

    // read files
    tsp = (float *) malloc(size * sizeof(float));
    retsize = fread(tsp, sizeof(float), size, fp);

    fclose(fp);
    fp = NULL;

    if (retsize != size) {
        harkio_Log_FileOpenFailed(path);
        free(tsp);
        return NULL;
    }

    *tsplen = size;
    return tsp;
}

/*------------------------------------------------------*/
// save float as 32bit float wav
/*------------------------------------------------------*/
int saveas(float *tsp, int tsplen, int samplingrate, const char *fname)
{
    SF_INFO sinfo;
    SNDFILE *sf;

    // setup header
    sinfo.frames = tsplen;
    sinfo.channels = 1;
    sinfo.samplerate = samplingrate;
    sinfo.format = SF_FORMAT_WAV|SF_FORMAT_FLOAT;

    // format check
    if (sf_format_check(&sinfo) == 0) {
        harkio_Log_printf(harkio_Log_Error, "Format Error.");
    }

    // write data
    sf = sf_open(fname, SFM_WRITE, &sinfo);
    sf_writef_float(sf, tsp, tsplen);
    sf_close(sf);
    return EXIT_SUCCESS;
}

int main(int argc, char *argv[])
{
    // arguments
    int numOfTSP = 16;
    int offset = 0;
    int samplingRate = 16000;
    char* tspPath;
    float ceilingAmp = -1.0;
    char* logLevel = "E";

    // tsp buffers
    float *origtsp, *gentsp;
    long tsplen;
    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "num", 'n',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &numOfTSP, 'n',
      "Number of TSPs", NULL },
    { "offset", 'o',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &offset, 'o',
      "Offset of the TSP", NULL },
    { "rate", 'r',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &samplingRate, 'r',
      "Sampling rate of TSP", NULL },
    { "src", 's',
      POPT_ARG_STRING,
      &tspPath, 's',
      "Input tsp raw (32 bit float) file", NULL },
    { "volume", 'v',
      POPT_ARG_FLOAT,
      &ceilingAmp, 'v',
      "Max amplitude for normalization", NULL },
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* output_filename");

    // argc check
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error, "argument error!!");
        help(optCon, EXIT_FAILURE);
    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case 'n':
            if(numOfTSP <= 0) {
                harkio_Log_printf(harkio_Log_Error, "Invalid '-n' value. Specify the integer (>= 1).");
                exit(EXIT_FAILURE);
            }
            hasParsedOpt = true;
            break;
        case 'o':
            if (offset < 0) {
                harkio_Log_printf(harkio_Log_Error, "Invalid '-o' value. Specify the integer (>= 0).");
                exit(EXIT_FAILURE);
            }
            hasParsedOpt = true;
            break;
        case 'r':
            if (samplingRate <= 0) {
                harkio_Log_printf(harkio_Log_Error, "Invalid '-r' value. Specify the integer (>= 1).");
                exit(EXIT_FAILURE);
            }
            hasParsedOpt = true;
            break;
        case 'v':
            if((ceilingAmp <= 0.0) || (ceilingAmp > 1.0)) {
                harkio_Log_printf(harkio_Log_Error, "Invalid '-v' value. Specify the real number (0.0 < value <= 1.0).");
                exit(EXIT_FAILURE);
            }
            hasParsedOpt = true;
            break;
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    // get filename
    const char* outfile = poptGetArg(optCon);

    // debug out
    harkio_Log_printf(harkio_Log_Debug, "n: %d", numOfTSP);
    harkio_Log_printf(harkio_Log_Debug, "o: %d", offset);
    harkio_Log_printf(harkio_Log_Debug, "r: %d", samplingRate);
    harkio_Log_printf(harkio_Log_Debug, "s: %s", tspPath);
    harkio_Log_printf(harkio_Log_Debug, "v: %f", ceilingAmp);

    // read original TSP
    if ((origtsp = readRawTSP(tspPath, &tsplen)) == NULL) {
        harkio_Log_printf(harkio_Log_Error, "Reading raw tsp failed.");
        free(origtsp);
        exit(EXIT_FAILURE);
    }

    // get tsp max amplitude
    float maxamp = 0.0;
    for (int i = 0; i < tsplen; i++) {
        float absamp = fabsf(origtsp[i]);
        if(maxamp < absamp) {
            maxamp = absamp;
        }
    }
    if (ceilingAmp == -1.0) {
        if(maxamp > 1.0) {
            fprintf(stderr, "Max amplitude is over 1.0. Output wave file may be clipped.\n");
        }
    } else {
        // normalize
        double ratio = 1.0 / maxamp * ceilingAmp;
        for (int i = 0; i < tsplen; i++) {
            origtsp[i]= (float)(origtsp[i] * ratio);
        }
    }

    // allocate memory for multiple TSP + offset
    gentsp = (float *) malloc(sizeof(float) * (tsplen * numOfTSP + offset));

    // add offset
    memset(gentsp, 0, sizeof(float) * offset);

    // copy tsps
    for (int i=0; i<numOfTSP; i++) {
        memcpy(gentsp + offset + tsplen * i,
               origtsp, sizeof(float) * tsplen);
    }

    if (!outfile) {
        harkio_Log_printf(harkio_Log_Error , "missing output_filename");
        exit(EXIT_FAILURE);
    }

    // save file
    if (saveas(gentsp, tsplen * numOfTSP + offset,
               samplingRate, outfile) == EXIT_FAILURE) {
        harkio_Log_printf(harkio_Log_Error, "Save file failed");
    }

    // free them
    free(origtsp);
    free(gentsp);

    harkio_Log_printf(harkio_Log_Debug, "File generated : %s", outfile);

    fprintf(stderr, "SUCCESS\n");
    fprintf(stderr, " output file : %s\n", outfile);

    // free poptContext
    poptFreeContext(optCon);

    exit(EXIT_SUCCESS);

}
